#!/usr/bin/perl
# $Id: starter_pdfvt1.pl,v 1.16.2.1 2013/07/05 10:42:43 rp Exp $
#
# Starter sample for PDF/VT-1
# Create a large number of invoices in a single PDF and make use of
# the following PDF/VT-1 features:
# - create a document part (DPart) hierarchy
# - assign PDF/VT scope attributes to images and imported PDF pages
# - add document part metadata (DPM) to the DPart root node and all page nodes
#
# Required software: PDFlib+PDI/PPS 9
# Required data: PDF background, fonts, several raster images

use PDFlib::PDFlib 9.0;
use strict;

use constant MAXRECORD => 100;

my	$i;
my	$stationery;
my	$page;
my      $record;
my      $barcodeimage;
my 	$stationeryfilename = "stationery_pdfx4p.pdf";
my 	$salesrepfilename = "sales_rep\%d.jpg";
my 	$fontname = "LinLibertine_R";

# This is where font/image/PDF input files live. Adjust as necessary.
my 	$searchpath = "../data";

my 	$left = 55;
my 	$right = 530;
my      $bottom = 822;

my	$fontsize = 12;
my 	$leading = $fontsize + 2;
my	$x;
my	$y;
my	$sum;
my	$total;
my	$buf;
my	$optlist;
my      $fontoptions;

my $closingtext =
    "Terms of payment: <save fillcolor={cmyk 0 1 1 0}>30 days net<restore>. " .
    "90 days warranty starting at the day of sale. " .
    "This warranty covers defects in workmanship only. " .
    "Kraxi Systems, Inc. will, at its option, repair or replace the " .
    "product under the warranty. This warranty is not transferable. " .
    "No returns or exchanges will be accepted for wet products.";

my @articledata = (
    { name=>"Super Kite",		price=>20 },
    { name=>"Turbo Flyer",		price=>40 },
    { name=>"Giga Trash",		price=>180 },
    { name=>"Bare Bone Kit",	        price=>50 },
    { name=>"Nitty Gritty",		price=>20 },
    { name=>"Pretty Dark Flyer",	price=>75 },
    { name=>"Large Hadron Glider",	price=>85 },
    { name=>"Flying Bat",         	price=>25 },
    { name=>"Simple Dimple",      	price=>40 },
    { name=>"Mega Sail",          	price=>95 },
    { name=>"Tiny Tin",           	price=>25 },
    { name=>"Monster Duck",      	price=>275 },
    { name=>"Free Gift",		price=>0 }
);

my @addressdata = (
    { firstname=>"Edith", lastname=>"Poulard", flat=>"Suite C",
			    street=>"Main Street", city=>"New York"},
    { firstname=>"Max", lastname=>"Huber", flat=>"",
			    street=>"Lipton Avenue", city=>"Albuquerque"},
    { firstname=>"Herbert", lastname=>"Pakard", flat=>"App. 29",
			    street=>"Easel", city=>"Duckberg" },
    { firstname=>"Charles", lastname=>"Fever", flat=>"Office 3",
			    street=>"Scenic Drive", city=>"Los Angeles" },
    { firstname=>"D.", lastname=>"Milliband", flat=>"",
			    street=>"Old Harbour", city=>"Westland" },
    { firstname=>"Lizzy", lastname=>"Tin", flat=>"Workshop",
			    street=>"Ford", "Detroit" },
    { firstname=>"Patrick", lastname=>"Black", flat=>"Backside",
			    street=>"Woolworth Street", city=>"Clover" },
);

my @salesrepnames = (
    "Charles Ragner",
    "Hugo Baldwin",
    "Katie Blomock",
    "Ernie Bastel",
    "Lucy Irwin",
    "Bob Montagnier",
    "Chuck Hope",
    "Pierre Richard"
);

my @salesrepimage;

my @headers = (
    "ITEM", "DESCRIPTION", "QUANTITY", "PRICE", "AMOUNT"
);

my @alignments = (
    "right", "left", "right", "right", "right"
);
        
my @months = (
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
);

my $dpm=0;
my $cip4_root;
my $cip4_metadata;

# Simulate a datamatrix barcode
my $MATRIXROWS          = 32;
my $MATRIXDATASIZE      = 4 * $MATRIXROWS;

sub
create_datamatrix
{
    my $record = @_;
    my @data;
    my $i;
    my $x;

    for ($i=0; $i<$MATRIXROWS; $i++)
    {
	$data[4*$i+0] = ((0xA3 + 1*$record + 17*$i) % 0xFF);
	$data[4*$i+1] = ((0xA2 + 3*$record + 11*$i) % 0xFF);
	$data[4*$i+2] = ((0xA0 + 5*$record +  7*$i) % 0xFF);
	$data[4*$i+3] = ((0x71 + 7*$record +  9*$i) % 0xFF);
    }
    for ($i=0; $i<$MATRIXROWS; $i++)
    {
	$data[4*$i+0] |= 0x80;
	$data[4*$i+2] |= 0x80;
	if ($i%2) {
	    $data[4*$i+3] |= 0x01;
	} else {
	    $data[4*$i+3] &= 0xFE;
	}
    }
    for ($i=0; $i<4; $i++)
    {
	$data[4*($MATRIXROWS/2-1)+$i] = 0xFF;
	$data[4*($MATRIXROWS-1)+$i] = 0xFF;
    }
    return pack("C$MATRIXDATASIZE", @data);
}

# create a new PDFlib object
my $p = new PDFlib::PDFlib;
eval {
    if ($p->begin_document("starter_pdfvt1.pdf",
	    "pdfx=PDF/X-4 pdfvt=PDF/VT-1 usestransparency=false " .
	    "nodenamelist={root recipient} recordlevel=1") == -1)
    {
	die("Error: " . $p->get_errmsg());
    }

    $p->set_option("SearchPath={{" . $searchpath ."}}");
    $p->set_option("errorpolicy=return");

    # all strings are expected as utf8
    $p->set_option("stringformat=utf8");

    $p->set_info("Creator", "PDFlib starter sample");
    $p->set_info("Title", "starter_pdfvt1");

    $fontoptions = "fontname=" . $fontname . " fontsize=" . $fontsize
                            . " embedding encoding=unicode";
                            
    # Define output intent profile
    if ($p->load_iccprofile("ISOcoated.icc", "usage=outputintent") == -1)
    {
	printf("Error: %s\n", $p->get_errmsg());
	die("Please install the ICC profile package from " .
	       "www.pdflib.com to run the PDF/VT-1 starter sample.\n");
    }

    # -----------------------------------
    # Load company stationery as background (used on first page
    # for each recipient)
    # -----------------------------------
    #/
    $stationery = $p->open_pdi_document($stationeryfilename, "");
    if ($stationery == -1) {
	die("Error: " . $p->get_errmsg());
    }

    $page = $p->open_pdi_page($stationery, 1,
	    "pdfvt={scope=global environment={Kraxi Systems}}");
    if ($page == -1) {
	die("Error: " . $p->get_errmsg());
    }

    # -----------------------------------
    # Preload images of all local sales reps (used on first page
    # for each recipient). To get encapsulated image XObjects,
    # the renderingintent option is used.
    # -----------------------------------
    #/
    for ($i=0; $i < scalar @salesrepnames; $i++)
    {
	$buf = sprintf($salesrepfilename, $i);
	$salesrepimage[$i] = $p->load_image("auto", $buf,
		"pdfvt={scope=file} renderingintent=Perceptual");

	if ($salesrepimage[$i] == -1) {
	    die("Error: " . $p->get_errmsg());
	}
    }

    # -----------------------------------
    # Construct DPM metadata for the DPart root node
    # -----------------------------------
    #/
    $cip4_metadata = $p->poca_new("containertype=dict usage=dpm");
    $p->poca_insert($cip4_metadata,
	    "type=string key=CIP4_Conformance value=base");
    $p->poca_insert($cip4_metadata,
	    "type=string key=CIP4_Creator value=starter_pdfvt1");
    $p->poca_insert($cip4_metadata,
	    "type=string key=CIP4_JobID value={Kraxi Systems invoice}");
	    
    $optlist = sprintf("containertype=dict usage=dpm " .
                        "type=dict key=CIP4_Metadata value=%d",
		        $cip4_metadata);
    $cip4_root = $p->poca_new($optlist);
	    
    $optlist = sprintf("containertype=dict usage=dpm " .
                        "type=dict key=CIP4_Root value=%d", $cip4_root);
    $dpm = $p->poca_new($optlist);

    # Create root node in the DPart hierarchy and add DPM metadata 
    $optlist = sprintf("dpm=%d", $dpm);
    $p->begin_dpart($optlist);

    $p->poca_delete($dpm, "recursive=true");

    for ($record=0; $record < MAXRECORD; $record++)
    {
	my $datamatrix;
	my $cip4_recipient;
	my $cip4_contact;
	my $cip4_person;
	my $firstname;
	my $lastname;
	my $result;
	my $pagecount = 0;
	my $item;

	$firstname = $addressdata[int(rand($#addressdata))]{firstname};
	$lastname = $addressdata[int(rand($#addressdata))]{lastname};

	# -----------------------------------
	# Construct DPM metadata for the next DPart node (i.e. the page)
	# -----------------------------------
	$dpm            = $p->poca_new("containertype=dict usage=dpm");
	$cip4_root      = $p->poca_new("containertype=dict usage=dpm");
	$cip4_recipient = $p->poca_new("containertype=dict usage=dpm");
	$cip4_contact   = $p->poca_new("containertype=dict usage=dpm");
	$cip4_person    = $p->poca_new("containertype=dict usage=dpm");

	$optlist = sprintf("type=dict key=CIP4_Root value=%d", $cip4_root);
	$p->poca_insert($dpm, $optlist);

	$optlist = sprintf("type=dict key=CIP4_Recipient value=%d",
						       $cip4_recipient);
	$p->poca_insert($cip4_root, $optlist);

	$optlist = sprintf("type=string key=CIP4_UniqueID value={ID_%d}",
			    $record);
	$p->poca_insert($cip4_recipient, $optlist);

	$optlist = sprintf("type=dict key=CIP4_Contact value=%d",
						       $cip4_contact);
	$p->poca_insert($cip4_recipient, $optlist);

	$optlist = sprintf("type=dict key=CIP4_Person value=%d",
						       $cip4_person);
	$p->poca_insert($cip4_contact, $optlist);

	$optlist = sprintf("type=string key=CIP4_Firstname value={%s}",
			    $firstname);
	$p->poca_insert($cip4_person, $optlist);

	$optlist = sprintf("type=string key=CIP4_Lastname value={%s}",
			    $lastname);
	$p->poca_insert($cip4_person, $optlist);

	# Create a new node in the document part hierarchy and
	# add DPM metadata
	$optlist = sprintf("dpm=%d", $dpm);
	$p->begin_dpart($optlist);

	$p->poca_delete($dpm, "recursive=true");

        # -----------------------------------
        # Create and place table with article list
        # -----------------------------------
        # 
        # ---------- Header row
        my $row = 1;
        my $tbl = -1;
        my $col;
    
        for ($col=1; $col <= scalar @headers; $col++)
        {
            $optlist = sprintf(
                "fittextline={position={%s center} %s} margin=2",
                $alignments[$col-1], $fontoptions);
            $tbl = $p->add_table_cell($tbl, $col, $row, $headers[$col-1],
                                        $optlist);
        }
        $row++;
    
        # ---------- Data rows: one for each article
        $total = 0;
    
        # -----------------------------------
        # Print variable-length article list
        # -----------------------------------
        for ($i = 0, $item = 0; $i < scalar @articledata; $i++) {
            my $quantity = int(rand(9)) + 1;
    
            if (int(rand(2)) % 2) {
                next;
            }
    
            $col = 1;
    
            $item++;
            my $sum = $articledata[$i]{price} * $quantity;
            
            # column 1: ITEM
            $buf = sprintf("%d", $item);
            $optlist = sprintf(
                "fittextline={position={%s center} %s} colwidth=5%% " .
                "margin=2",
                $alignments[$col-1], $fontoptions);
            $tbl = $p->add_table_cell($tbl, $col++, $row, $buf, $optlist);
    
            # column 2: DESCRIPTION
            $optlist = sprintf(
                "fittextline={position={%s center} %s} colwidth=50%% " .
                "margin=2",
                $alignments[$col-1], $fontoptions);
            $tbl = $p->add_table_cell($tbl, $col++, $row,
                    $articledata[$i]{name}, $optlist);
    
            # column 3: QUANTITY
            $buf = sprintf("%d", $quantity);
            $optlist = sprintf(
                "fittextline={position={%s center} %s} margin=2",
                $alignments[$col-1], $fontoptions);
            $tbl = $p->add_table_cell($tbl, $col++, $row, $buf, $optlist);
    
            # column 4: PRICE
            $buf = sprintf("%.2f", $articledata[$i]{price});
            $optlist = sprintf(
                "fittextline={position={%s center} %s} margin=2",
                $alignments[$col-1], $fontoptions);
            $tbl = $p->add_table_cell($tbl, $col++, $row, $buf, $optlist);
    
            # column 5: AMOUNT
            $buf = sprintf("%.2f", $sum);
            $optlist = sprintf(
                "fittextline={position={%s center} %s} margin=2",
                $alignments[$col-1], $fontoptions);
            $tbl = $p->add_table_cell($tbl, $col++, $row, $buf, $optlist);
    
            $total += $sum;
            $row++;
        }
    
        # ---------- Print total in the rightmost column
        $buf = sprintf("%.2f", $total);
        $optlist = sprintf(
            "fittextline={position={%s center} %s} margin=2",
            $alignments[$#alignments], $fontoptions);
        $tbl = $p->add_table_cell($tbl, scalar @headers, $row++, $buf,
                                    $optlist);
    
    
        # ---------- Footer row with terms of payment
        $optlist = sprintf("%s alignment=justify leading=120%%", $fontoptions);
        my $tf = $p->create_textflow($closingtext, $optlist);
    
        $optlist = sprintf(
                "rowheight=1 margin=2 margintop=%f textflow=%d colspan=%d",
                2*$fontsize, $tf, scalar @headers);
        $tbl = $p->add_table_cell($tbl, 1, $row++, "", $optlist);
    
    
        # ----- Place the table instance(s), creating pages as required
        do {
            my $top;
    
            $p->begin_page_ext(0, 0,
                    "topdown=true width=a4.width height=a4.height");
    
            if (++$pagecount == 1)
            {
                # -----------------------------------
                # Place company stationery as background on first page
                # for each recipient
                # -----------------------------------
                $p->fit_pdi_page($page, 0, 842, "");
    
                # -----------------------------------
                # Place name and image of local sales rep on first page
                # for each recipient
                # -----------------------------------
                $y = 177;
                $x = 455;
    
                $optlist = sprintf(
                        "fontname=%s encoding=winansi embedding fontsize=9",
                        $fontname);
                $p->fit_textline("Local sales rep:", $x, $y, $optlist);
                $p->fit_textline($salesrepnames[$record % scalar @salesrepnames],
                        $x, $y+9, $optlist);
    
                $y = 280;
                $p->fit_image($salesrepimage[$record % scalar @salesrepnames], $x, $y,
                        "boxsize={90 90} fitmethod=meet");
    
    
                # -----------------------------------
                # Address of recipient
                # -----------------------------------
                $y = 170;
    
                $optlist = sprintf(
                    "fontname=%s encoding=winansi embedding fontsize=%f",
                    $fontname, $fontsize);
                $buf = sprintf("%s %s", $firstname, $lastname);
                $p->fit_textline($buf, $left, $y, $optlist);
    
                $y += $leading;
                $p->fit_textline(
                        $addressdata[int(rand(scalar @addressdata))]{flat},
                        $left, $y, $optlist);
    
                $y += $leading;
                $buf = sprintf("%d %s",
                        rand(999),
                        $addressdata[int(rand(scalar @addressdata))]{street});
                $p->fit_textline($buf, $left, $y, $optlist);
    
                $y += $leading;
                $buf = sprintf("%05d %s",
                        int(rand(99999)),
                        $addressdata[int(rand(scalar @addressdata))]{city});
                $p->fit_textline($buf, $left, $y, $optlist);
    
    
                # -----------------------------------
                # Individual barcode image for each recipient. To get
                # encapsulated image XObjects the renderingintent option
                #  is used.
                # -----------------------------------
                $datamatrix = create_datamatrix($record);
                $p->create_pvf("barcode", $datamatrix, "");
    
                $barcodeimage = $p->load_image("raw", "barcode",
                    "bpc=1 components=1 width=32 height=32 invert " .
                    "pdfvt={scope=singleuse} renderingintent=Saturation");
                if ($barcodeimage == -1) {
                    die("Error: " . $p->get_errmsg());
                }
    
                $p->fit_image($barcodeimage, 280.0, 200.0, "scale=1.5");
                $p->close_image($barcodeimage);
                $p->delete_pvf("barcode");
    
    
                # -----------------------------------
                # Print header and date
                # -----------------------------------
                $y = 300;
                $buf = sprintf("INVOICE %d-%d", (localtime)[5]+1900, $record+1);
                $optlist = sprintf(
                    "fontname=%s encoding=winansi embedding fontsize=%d",
                    $fontname, $fontsize);
                $p->fit_textline($buf, $left, $y, $optlist);
    
                $buf = sprintf("%s %d, %d", $months[(localtime)[4]],
                             (localtime)[3], (localtime)[5]+1900);
                $optlist = sprintf(
                    "fontname=%s encoding=unicode fontsize=%d embedding " .
                    "position {100 0}", $fontname, $fontsize);
                $p->fit_textline($buf, $right, $y, $optlist);
    
                $top = $y + 2*$leading;
            }
            else
            {
                $top = 50;
            }
    
            # Place the table on the page.
            # Shade every other row, except the footer row.
            $result = $p->fit_table($tbl,
                    $left, $bottom, $right, $top,
                    "header=1 " .
                    "fill={{area=rowodd fillcolor={gray 0.9}} " .
                        "{area=rowlast fillcolor={gray 1}}} " .
                    "rowheightdefault=auto colwidthdefault=auto");
    
            if ($result eq "_error")
            {
                die("Error when placing table: " . $p->get_errmsg());
            }
    
            $p->end_page_ext("");
        } while ($result eq "_boxfull");
    
        $p->delete_table($tbl, "");

	# Close node in the document part hierarchy
	$p->end_dpart("");
    }

    $p->close_pdi_page($page);
    $p->close_pdi_document($stationery);

    for ($i = 0; $i < scalar @salesrepimage; $i++)
    {
	$p->close_image($salesrepimage[$i]);
    }

    # Close root node in the document part hierarchy
    $p->end_dpart("");

    $p->end_document("");
};

if ($@) {
    die("$0: PDFlib Exception occurred:\n$@");
}
